/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __spreadArrays() {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
}

var GunEvent = /** @class */ (function () {
    function GunEvent(name) {
        if (name === void 0) { name = "GunEvent"; }
        this.name = name;
        this._listeners = [];
        this.listenerCount = this.listenerCount.bind(this);
        this.on = this.on.bind(this);
        this.off = this.off.bind(this);
        this.trigger = this.trigger.bind(this);
    }
    GunEvent.prototype.listenerCount = function () {
        return this._listeners.length;
    };
    GunEvent.prototype.on = function (cb) {
        if (this._listeners.indexOf(cb) !== -1)
            return;
        this._listeners.push(cb);
    };
    GunEvent.prototype.off = function (cb) {
        var idx = this._listeners.indexOf(cb);
        if (idx !== -1)
            this._listeners.splice(idx, 1);
    };
    GunEvent.prototype.reset = function () {
        this._listeners = [];
    };
    GunEvent.prototype.trigger = function (a, b, c) {
        this._listeners.forEach(function (cb) { return cb(a, b, c); });
    };
    return GunEvent;
}());

var ChainGunLink = /** @class */ (function () {
    function ChainGunLink(chain, key, parent) {
        this.key = key;
        this._chain = chain;
        this._parent = parent;
        this._hasReceived = false;
        this._updateEvent = new GunEvent(this.getPath().join("|"));
    }
    ChainGunLink.prototype.getPath = function () {
        if (this._parent)
            return __spreadArrays(this._parent.getPath(), [this.key]);
        return [this.key];
    };
    /**
     * Traverse a location in the graph
     *
     * @param key Key to read data from
     * @param cb
     * @returns New chain context corresponding to given key
     */
    ChainGunLink.prototype.get = function (key, cb) {
        return new this.constructor(this._chain, key, this);
    };
    /**
     * Move up to the parent context on the chain.
     *
     * Every time a new chain is created, a reference to the old context is kept to go back to.
     *
     * @param amount The number of times you want to go back up the chain. {-1} or {Infinity} will take you to the root.
     * @returns a parent chain context
     */
    ChainGunLink.prototype.back = function (amount) {
        if (amount === void 0) { amount = 1; }
        if (amount < 0 || amount === Infinity)
            return this._chain;
        if (amount === 1)
            return this._parent || this._chain;
        return this.back(amount - 1);
    };
    /**
     * Save data into gun, syncing it with your connected peers.
     *
     * You do not need to re-save the entire object every time, gun will automatically
     * merge your data into what already exists as a "partial" update.
     *
     * @param value the data to save
     * @param cb an optional callback, invoked on each acknowledgment
     * @returns same chain context
     */
    ChainGunLink.prototype.put = function (value, cb) {
        throw new Error("put() isn't supported yet");
        return this;
    };
    /**
     * Add a unique item to an unordered list.
     *
     * Works like a mathematical set, where each item in the list is unique.
     * If the item is added twice, it will be merged.
     * This means only objects, for now, are supported.
     *
     * @param data should be a gun reference or an object
     * @param cb The callback is invoked exactly the same as .put
     * @returns chain context for added object
     */
    ChainGunLink.prototype.set = function (data, cb) {
        throw new Error("set() isn't supported yet");
    };
    /**
     * Register a callback for when it appears a record does not exist
     *
     * If you need to know whether a property or key exists, you can check with .not.
     * It will consult the connected peers and invoke the callback if there's reasonable certainty that none of them have the data available.
     *
     * @param cb If there's reason to believe the data doesn't exist, the callback will be invoked. This can be used as a check to prevent implicitly writing data
     * @returns same chain context
     */
    ChainGunLink.prototype.not = function (cb) {
        var _this = this;
        this.on(function (val) { return typeof val === "undefined" && cb(_this.key); });
        return this;
    };
    /**
     * Change the configuration of this chain link
     *
     * @param options
     * @returns same chain context
     */
    ChainGunLink.prototype.opt = function (options) {
        return this;
    };
    /**
     * Get the current data without subscribing to updates. Or undefined if it cannot be found.
     *
     * @param cb The data is the value for that chain at that given point in time. And the key is the last property name or ID of the node.
     * @returns same chain context
     */
    ChainGunLink.prototype.once = function (cb) {
        var _this = this;
        this.promise().then(function (val) { return cb(val, _this.key); });
        return this;
    };
    /**
     * Subscribe to updates and changes on a node or property in realtime.
     *
     * Triggered once initially and whenever the property or node you're focused on changes,
     * Since gun streams data, the callback will probably be called multiple times as new chunk comes in.
     *
     * To remove a listener call .off() on the same property or node.
     *
     * @param cb The callback is immediately fired with the data as it is at that point in time.
     * @returns same chain context
     */
    ChainGunLink.prototype.on = function (cb) {
        this._updateEvent.on(cb);
        if (!this._endQuery) {
            this._endQuery = this._chain.query(this.getPath(), this._onQueryResponse.bind(this));
        }
        if (this._hasReceived)
            cb(this._lastValue, this.key);
        return this;
    };
    /**
     * Unsubscribe one or all listeners subscribed with on
     *
     * @returns same chain context
     */
    ChainGunLink.prototype.off = function (cb) {
        if (cb) {
            this._updateEvent.off(cb);
            if (this._endQuery && !this._updateEvent.listenerCount()) {
                this._endQuery();
            }
        }
        else {
            if (this._endQuery)
                this._endQuery();
            this._updateEvent.reset();
        }
        return this;
    };
    ChainGunLink.prototype.promise = function () {
        var _this = this;
        return new Promise(function (ok) {
            var cb = function (val) {
                ok(val);
                _this.off(cb);
            };
            _this.on(cb);
        });
    };
    ChainGunLink.prototype.then = function (fn) {
        return this.promise().then(fn);
    };
    /**
     * Iterates over each property and item on a node, passing it down the chain
     *
     * Not yet supported
     *
     * Behaves like a forEach on your data.
     * It also subscribes to every item as well and listens for newly inserted items.
     *
     * @returns a new chain context holding many chains simultaneously.
     */
    ChainGunLink.prototype.map = function () {
        throw new Error("map() isn't supported yet");
        return this.get(""); // Special case key
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.path = function (path) {
        throw new Error("No plans to support this");
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.open = function (cb) {
        throw new Error("No plans to support this");
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.load = function (cb) {
        throw new Error("No plans to support this");
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.bye = function () {
        throw new Error("No plans to support this");
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.later = function () {
        throw new Error("No plans to support this");
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.unset = function (node) {
        throw new Error("No plans to support this");
    };
    ChainGunLink.prototype._onQueryResponse = function (value) {
        this._updateEvent.trigger(value, this.key);
        this._lastValue = value;
        this._hasReceived = true;
    };
    return ChainGunLink;
}());

var DEFAULT_OPTS = {
    futureGrace: 10 * 60 * 1000,
    Lexical: JSON.stringify // what gun.js uses
};
var EMPTY = {};
function diffGunCRDT(existingGraph, updatedGraph, opts) {
    if (opts === void 0) { opts = DEFAULT_OPTS; }
    var _a = opts || EMPTY, _b = _a.machineState, machineState = _b === void 0 ? new Date().getTime() : _b, _c = _a.futureGrace, futureGrace = _c === void 0 ? DEFAULT_OPTS.futureGrace : _c, _d = _a.Lexical, Lexical = _d === void 0 ? DEFAULT_OPTS.Lexical : _d;
    var maxState = machineState + futureGrace; // eslint-disable-line
    var allUpdates = {};
    for (var soul in updatedGraph) {
        var existing = existingGraph[soul];
        var updated = updatedGraph[soul];
        var existingState = (existing && existing._ && existing._['>']) || EMPTY;
        var updatedState = (updated && updated._ && updated._['>']) || EMPTY;
        if (!updated) {
            if (!(soul in existingGraph))
                allUpdates[soul] = updated;
            continue;
        }
        var hasUpdates = false;
        var updates = {
            _: {
                '#': updated._['#'],
                '>': {}
            }
        };
        for (var key in updatedState) {
            var existingKeyState = existingState[key];
            var updatedKeyState = updatedState[key];
            if (updatedKeyState > maxState || !updatedKeyState)
                continue;
            if (existingKeyState && existingKeyState >= updatedKeyState)
                continue;
            if (existingKeyState === updatedKeyState) {
                var existingVal = (existing && existing[key]) || undefined;
                var updatedVal = updated[key];
                // This is based on Gun's logic
                if (Lexical(updatedVal) <= Lexical(existingVal))
                    continue;
            }
            updates[key] = updated[key];
            updates._['>'][key] = updatedKeyState;
            hasUpdates = true;
        }
        if (hasUpdates) {
            allUpdates[soul] = updates;
        }
    }
    return Object.keys(allUpdates) ? allUpdates : undefined;
}

function mergeGunNodes(existing, updates) {
    if (!existing)
        return updates;
    if (!updates)
        return existing;
    var existingMeta = existing._;
    return __assign(__assign(__assign({}, existing), updates), { _: {
            '#': existingMeta['#'],
            '>': __assign(__assign({}, existingMeta['>']), updates._['>'])
        } });
}

function diffSets(initial, updated) {
    return [
        updated.filter(function (key) { return initial.indexOf(key) === -1; }),
        initial.filter(function (key) { return updated.indexOf(key) === -1; })
    ];
}

var ChainGun = /** @class */ (function () {
    function ChainGun(LinkClass) {
        if (LinkClass === void 0) { LinkClass = ChainGunLink; }
        this._opt = {};
        this._graph = {};
        this._listeners = {};
        this._readMiddleware = [];
        this._writeMiddleware = [];
        this.LinkClass = LinkClass;
    }
    ChainGun.prototype.opt = function (options) {
        this._opt = __assign(__assign({}, this._opt), options);
        return this;
    };
    /**
     * Register graph middleware
     *
     * @param middleware The middleware function to add
     * @param kind Optionaly register write middleware instead of read by passing "write"
     */
    ChainGun.prototype.use = function (middleware, kind) {
        if (kind === void 0) { kind = "read"; }
        if (kind === "read") {
            this._readMiddleware.push(middleware);
        }
        else if (kind === "write") {
            this._writeMiddleware.push(middleware);
        }
        return this;
    };
    /**
     * Traverse a location in the graph
     *
     * @param key Key to read data from
     * @param cb
     * @returns New chain context corresponding to given key
     */
    ChainGun.prototype.get = function (soul, cb) {
        return new this.LinkClass(this, soul);
    };
    /**
     * Write node data
     *
     * @param data one or more gun nodes keyed by soul
     */
    ChainGun.prototype.put = function (data) {
        throw new Error("put() isn't supported yet");
        return this;
    };
    /**
     * Update node data in this chain from some local or external source
     *
     * @param soul soul of node to update
     * @param data node data to include
     */
    ChainGun.prototype.addNodeData = function (data) {
        return __awaiter(this, void 0, void 0, function () {
            var diff, i, souls, i, soul, listener, node;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        diff = data ? diffGunCRDT(this._graph, data) : undefined;
                        i = 0;
                        _a.label = 1;
                    case 1:
                        if (!(i < this._readMiddleware.length)) return [3 /*break*/, 4];
                        if (!diff)
                            return [2 /*return*/];
                        return [4 /*yield*/, this._readMiddleware[i](diff, this._graph)];
                    case 2:
                        diff = _a.sent();
                        _a.label = 3;
                    case 3:
                        i++;
                        return [3 /*break*/, 1];
                    case 4:
                        if (!diff)
                            return [2 /*return*/];
                        souls = Object.keys(diff);
                        for (i = 0; i < souls.length; i++) {
                            soul = souls[i];
                            listener = this._listeners[soul];
                            if (!listener) {
                                this._forget(soul);
                                continue;
                            }
                            node = (this._graph[soul] = mergeGunNodes(this._graph[soul], diff[soul]));
                            listener.trigger(node, soul);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    ChainGun.prototype.query = function (path, cb) {
        var _this = this;
        var lastSouls = [];
        var currentValue;
        var updateQuery = function () {
            var _a = _this._getPathData(path), souls = _a.souls, value = _a.value, complete = _a.complete;
            var _b = diffSets(lastSouls, souls), added = _b[0], removed = _b[1];
            if ((complete && typeof currentValue === "undefined") ||
                (typeof value !== "undefined" && value !== currentValue)) {
                currentValue = value;
                cb(value, path[path.length - 1]);
            }
            added.forEach(function (soul) { return _this._request(soul, updateQuery); });
            removed.forEach(function (soul) { return _this._unlisten(soul, updateQuery); });
        };
        updateQuery();
        return function () {
            lastSouls.forEach(function (soul) { return _this._unlisten(soul, updateQuery); });
        };
    };
    ChainGun.prototype._request = function (soul, cb) {
        if (!this._opt.askNodes)
            return;
        this._listen(soul, cb);
        this._opt.askNodes([soul]);
    };
    ChainGun.prototype._listen = function (soul, cb) {
        var _this = this;
        var listener = this._listeners[soul];
        if (!listener) {
            listener = this._listeners[soul] = new GunEvent(soul);
        }
        listener.on(cb);
        return function () { return _this._unlisten(soul, cb); };
    };
    ChainGun.prototype._unlisten = function (soul, cb) {
        var listener = this._listeners[soul];
        if (listener)
            listener.off(cb);
        if (listener.listenerCount() <= 0)
            this._forget(soul);
    };
    ChainGun.prototype._getPathData = function (keys) {
        var lastKey = keys[keys.length - 1];
        if (keys.length === 1) {
            return {
                souls: keys,
                value: this._graph[lastKey],
                complete: lastKey in this._graph
            };
        }
        var _a = this._getPathData(keys.slice(0, keys.length - 1)), parentValue = _a.value, souls = _a.souls;
        if (typeof parentValue !== "object") {
            return {
                souls: souls,
                value: undefined,
                complete: typeof parentValue !== "undefined"
            };
        }
        var value = parentValue[lastKey];
        if (!value) {
            return {
                souls: souls,
                value: value,
                complete: true
            };
        }
        var edgeSoul = value["#"];
        if (edgeSoul) {
            return {
                souls: __spreadArrays(souls, [edgeSoul]),
                value: this._graph[edgeSoul],
                complete: edgeSoul in this._graph
            };
        }
        return {
            souls: souls,
            value: value,
            complete: true
        };
    };
    ChainGun.prototype._forget = function (soul) {
        if (this._listeners[soul]) {
            this._listeners[soul].reset();
            delete this._listeners[soul];
        }
        delete this._graph[soul];
        return this;
    };
    return ChainGun;
}());

function generateMessageId() {
    return Math.random()
        .toString(36)
        .slice(2);
}

function attachToGun(gunDb, chain) {
    var requests = {};
    var requestsBySoul = {};
    function askNodes(souls) {
        for (var i = 0; i < souls.length; i++) {
            var soul = souls[i];
            if (soul in requestsBySoul)
                continue;
            var msgId = generateMessageId();
            requests[msgId] = soul;
            requestsBySoul[soul] = msgId;
            gunDb.on("in", {
                "#": msgId,
                get: {
                    "#": soul
                }
            });
        }
    }
    gunDb.on("put", function (request) {
        var _a;
        this.to.next(request);
        if (!request)
            return;
        var respondingTo = request["@"];
        var requestedSoul = respondingTo && requests[respondingTo];
        if (request.put) {
            chain.addNodeData(request.put);
        }
        else if (requestedSoul) {
            chain.addNodeData((_a = {},
                _a[requestedSoul] = undefined,
                _a));
        }
    });
    chain.opt({ askNodes: askNodes });
    gunDb.chaingun = chain;
    return gunDb;
}

export { ChainGun, ChainGunLink, GunEvent, attachToGun, diffGunCRDT, generateMessageId, mergeGunNodes };
//# sourceMappingURL=chaingun.es5.js.map
