import { ChainGunLink } from "./ChainGunLink";
interface ChainGunOptions {
    askNodes?: (souls: string[]) => void;
}
export declare class ChainGun {
    private LinkClass;
    private _opt;
    private _readMiddleware;
    private _writeMiddleware;
    private _graph;
    private _listeners;
    constructor(LinkClass?: typeof ChainGunLink);
    opt(options: ChainGunOptions): this;
    /**
     * Register graph middleware
     *
     * @param middleware The middleware function to add
     * @param kind Optionaly register write middleware instead of read by passing "write"
     */
    use(middleware: ChainGunMiddleware, kind?: ChainGunMiddlewareType): this;
    /**
     * Traverse a location in the graph
     *
     * @param key Key to read data from
     * @param cb
     * @returns New chain context corresponding to given key
     */
    get(soul: string, cb?: GunPutCb): ChainGunLink;
    /**
     * Write node data
     *
     * @param data one or more gun nodes keyed by soul
     */
    put(data: {
        [soul: string]: GunNode;
    }): this;
    /**
     * Update node data in this chain from some local or external source
     *
     * @param soul soul of node to update
     * @param data node data to include
     */
    addNodeData(data: GunGraph): Promise<void>;
    query(path: string[], cb: GunOnCb): () => void;
    private _request;
    private _listen;
    private _unlisten;
    private _getPathData;
    private _forget;
}
export {};
